/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: asm1825.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 */

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>
#include "conf.h"
#include "util.h"
#include "memmap.h"
#include "macro.h"
#include "source.h"
#include "opcodes.h"
#include "symtab.h"
#include "eval.h"
#include "config.h"

const char rcsid[] = "$Id: asm1825.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $";

mmhandle chipmem;


struct sfstack {
  struct sfstack *next;
  FILE *fp;
};

struct sfstack *source_stack = NULL;

// int bsr = -1;

void strreplace(char *s, const char *m, const char *r)
{
  char *loc;
  int shift;
  int src_shift = 0;
  int i;

  message(MSG_INFO, "strreplace:  in: %s in %s", m, s);
  for (loc = strstr(s, m); loc != NULL; loc = strstr(s, m) ) {
    shift = (strlen(r)-strlen(m));
    if (shift < 0) { shift = 0; src_shift = 1; }
    message(MSG_DEBUG, "       s: %s", s);
    message(MSG_DEBUG, "  loc(1): %s, shift=%d, src_shift=%d",
        loc, shift, src_shift);
    message(MSG_DEBUG, "  loc(2): dest = '%s', src = '%s', n = %d",
		loc + shift, loc + src_shift, strlen(loc)+1);
    memmove((loc + shift), loc + src_shift, strlen(loc)+1);
    message(MSG_DEBUG, "  loc(3): %s", loc);
    i = 0;
    while (r[i] != '\0') {
      loc[i] = r[i];
      i++;
    }
  }
  message(MSG_INFO, "strreplace: out: %s in %s", m, s);
}


int splitargs(char *argstr, char *argv[])
{
  char *args = argstr;	// Don't modify argstr?
  int argc = 0;

  while (*args != '\0') {
    argv[argc++] = args;
    while ((*args != ',') && (*args != '\0')) args++;
    if (*args == ',') *(args++) = '\0';
  }
  argv[argc] = NULL;
  return argc;
}


void
do_org(const char *o)
{
  int v;

  if ((v = evalString(o)) == EVAL_FAIL) {
    // Label?
    message(MSG_WARNING, "ORG parameter of '%s' not understood", o);
  }

  message(MSG_INFO, "Setting LC to 0x%04x", v);
  lc = v;
  bsr = -1;	// Invalidate BSR tracking
}




void do_banksel(const char *op)
{
  unsigned short f = evalString(op);
  unsigned short fl = f & 0x007f;
  unsigned short cval = (f >> 7);
  char buf[8];

  /* Core registers and common RAM are bank independent */
  if ((fl <= 0x0b) || (fl >= 0x70)) return;

  if ((bsr < 0) || (bsr != cval)) {
    snprintf(buf, sizeof(buf), "0x%x", cval);
    try_opcode("movlb", buf, "");
    // bsr = cval;
  }
}



void do_config(int word, const char *op)
{
  unsigned short cval;

  cval = symResolveLiteral(op);
  mmWrite(chipmem, 0x8006 + word, cval);
}


void do_equ(const char *label, const char *op)
{
  int i;
  // sscanf(op, "%i", &i);
  i = evalString(op);
  symAdd(label, i, SF_ABS);
}


FILE * do_include(FILE *sfp, const char *op)
{
  struct sfstack *sp;
  FILE *fp;
  char pstr[MAXPATHLEN];

  // First, push the current file pointer onto the stack.
  sp = (struct sfstack *)malloc(sizeof(struct sfstack));
  sp->fp = sfp;
  sp->next = source_stack;
  source_stack = sp;

  // Next, open the new file.
  // Try current directory first.
  if ((fp = fopen(op, "r")) == NULL) {
    if (op[0] != '/') {
      snprintf(pstr, sizeof(pstr), "%s/%s", ASMINCLUDE_PATH, op);
      if ((fp = fopen(pstr, "r")) == NULL) {
        message(MSG_ERROR, "include file '%s' not found.\n", op);
        exit(1);
      }
    } else {
      message(MSG_ERROR, "include file '%s' not found.\n", op);
      exit(1);
    }
  }

  // And return the new file pointer.
  return fp;
}


// Process (insert) a defined macro.  May be called recursively
// from itself for macro calls within a macro.
//
void do_macro(const char *mname, char *params)
{
  StateList *sl;
  char label[LABEL_LEN];
  char opcode[32];
  char operands[64];
  char comment[256];
  char *argv[10];
  int argc;
  int inv;
  int i;
  char mark[4];

  sl = macLookup(mname);
  inv = macInvocations(mname, 1);
  // fprintf(stderr, "Calling splitargs(\"%s\", ...)\n", params);
  argc = splitargs(params, argv);
  message(MSG_INFO, "Macro %s called with %d arguments", mname, argc);
  // for (i=0; i<argc; i++) fprintf(stderr, "  %d: %s\n", i, argv[i]);
  while (sl != NULL) {
    stUse(&sl, label, opcode, operands, comment);
    if (label[0] == '@') sprintf(&(label[strlen(label)]), "%d", inv);
    for (i=0; i<argc; i++) {
      sprintf(mark, "$%d", i+1);
      strreplace(operands, mark, argv[i]);
      // This isn't perfect; works only when the local label is the
      // only operand.  But it should make local labels work right in
      // most cases.
      if (operands[0] == '@') sprintf(&(operands[strlen(operands)]), "%d", inv);
    }
    if (pass == 1)
        message(MSG_INFO, "%04x\t%s\t%s %s\t%s", lc, label, opcode, operands, comment);

    // Macro local labels and parameter substitions done. Now process the line.
    //

    // Add any label to the symbol list.
    //
    if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {
      if (pass == 0) {
        // Pass 1
        // Don't add the symbol if the opcode is the EQU directive
        if (strcmp(opcode, "equ")) symAdd(label, lc, 0);
      } else {
        // Pass 2
        // This should never be true unless there is a bug in the assember
        //
        if (strcmp(opcode, "equ") && (symLookup(label) != lc)) {
          message(MSG_CRITICAL, "Second pass value mismatch on label '%s'",
              label);
        }
      }  // if (pass)
    }

    // Process the opcode, if one exists
    //
    if (opcode[0] != '\0') {
      // fprintf(stderr, "DEBUG: processing opcode '%s'\n", opcode);
      try_opcode(opcode, operands, comment);
      if (!strcmp(opcode, "banksel")) do_banksel(operands);
      if (macInvocations(opcode, 0) >= 0) do_macro(opcode, operands);
    }

  }
}


FILE * pop_source(FILE *sfp)
{
  struct sfstack *sp;
  FILE *fp;

  // Close the current file.
  fclose(sfp);

  // If the stack is empty, return NULL. This means the main
  // source file has been closed.
  //
  if (!source_stack) return NULL;

  // Otherwise pop the last file pointer off the stack.
  fp = source_stack->fp;
  sp = source_stack;
  source_stack = source_stack->next;
  free(sp);

  // And return the pointer.
  return fp;
}


void pass1(const char *sf)
{
  FILE *sfp;
  char ibuf[512];
  char label[LABEL_LEN];
  char opcode[32];
  char operands[128];
  char comment[256];
  char macname[32];
  StateList **msl = NULL;

  // And make sure location counter is reset to start.
  lc = 0;
  pass = 0;
  macname[0] = '\0';	// Not in a macro define at start

  message(MSG_NOTICE, "Beginning pass 1");

  if ((sfp = fopen(sf, "r")) == NULL) {
    message(MSG_ERROR, "Error opening source file.");
    exit(1);
  }

  symInit();

  do {
    while (fgets(ibuf, sizeof(ibuf)-1, sfp) != NULL) {
      if (!srcParseLine(ibuf, label, opcode, operands, comment)) continue;
      lowercase(opcode);
      if (macname[0] != '\0') {
        // Reading and storing macro
        if (!strcmp(opcode, "endm")) {
          // fprintf(stderr, "Ending macro definition '%s'\n", macname);
          macname[0] = '\0';
        } else {
          // Store this line in the macro
          stAdd(msl, label, opcode, operands, comment);
        }
      } else {
        if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {
          // Don't add the symbol if the opcode is the EQU directive
          if (strcmp(opcode, "equ")) {
            symAdd(label, lc, 0);
            bsr = -1;  // Invalidate BSR tracking at label
          }
        }

        if (opcode[0] != '\0') {
          message(MSG_DEBUG, "processing opcode '%s'", opcode);
          try_opcode(opcode, operands, comment);

          // Right now we're just invalidating BSR tracking if the
          //   opcode was used.  We could read the operand
          //   and update 'bsr' accordingly.
          // movlb checking now takes place in opcodes.c
          // If everything checks out, this can be removed here.
          // if (!strcmp(opcode, "movlb")) bsr = -1;

          if (!strcmp(opcode, "end")) {
            message(MSG_INFO, "END statement reached.");
            // Close all input files
            while ((sfp = pop_source(sfp)));
            // Break out of the loops
            return;
          }
          if (!strcmp(opcode, "macro")) {
            message(MSG_INFO, "Defining macro '%s'", operands);
            strncpy(macname, operands, 32);
            msl = macAdd(macname, 0);
          }
          if (!strcmp(opcode, "banksel")) do_banksel(operands);
          if (!strcmp(opcode, "endm")) macname[0] = '\0';
          if (!strcmp(opcode, "equ")) do_equ(label, operands);
          if (!strcmp(opcode, "include")) sfp = do_include(sfp, operands);
          if (!strcmp(opcode, "org")) do_org(operands);
          if (macInvocations(opcode, 0) >= 0) do_macro(opcode, operands);
        }  // opcode not empty
      }  // Not in macro
    }
  } while ((sfp = pop_source(sfp)));

  // The last sfp will have been closed by pop_source
}




void
getroot(char *r, const char *f)
{
  int i;

  strncpy(r, f, 1023);
  i = strlen(r);
  while ((r[i] != '.') && (i > 0)) i--;
  if (r[i] == '.') r[i] = '\0';
}


int main(int argc, char *argv[])
{
  int opt;
  FILE *sfp;
  FILE *symp;
  char *src;
  char fnroot[MAXPATHLEN];
  char ibuf[512];
  char label[LABEL_LEN];
  char opcode[32];
  char operands[128];
  char comment[256];
  int voc;
  int done = 0;
  int inmacro = 0;

  lc = 0;
  pass = 0;
  while ((opt = getopt(argc, argv, "qv")) != -1) {
    switch (opt) {
      case 'q':
        msg_level--;
        break;
      case 'v':
        msg_level++;
        break;
      default:
        fprintf(stderr, "Usage: %s [ -qv ] <source>\n", argv[0]);
        exit(1);
    }
  }
  if (argv[optind] == NULL) {
    fprintf(stderr, "Usage: %s [ -qv ] <source>\n", argv[0]);
    exit(1);
  }

  src = argv[optind];

  message(MSG_INFO, "%s %s starting", PACKAGE, VERSION);

  getroot(fnroot, src);
  message(MSG_INFO, "Using root name: '%s'\n", fnroot);

  chipmem = mmInit(8192, 0x0b, 256, 0x8000);
  ocitInit();
  macInit();

  pass1(src);
  // macDump(stdout);

  lc = 0;
  pass = 1;
  message(MSG_NOTICE, "Beginning pass 2");
  macReset();

  sprintf(ibuf, "%s.lst", fnroot);
  if ((lstp = fopen(ibuf, "w")) == NULL) {
    message(MSG_WARNING, "Could not open list file for writing");
  }

  if ((sfp = fopen(src, "r")) == NULL) {
    message(MSG_ERROR, "Error opening source file.");
    exit(1);
  }

  // And make sure location counter is reset to start.
  lc = 0;

  do {
    while (fgets(ibuf, sizeof(ibuf)-1, sfp) != NULL) {
      if (!srcParseLine(ibuf, label, opcode, operands, comment)) continue;
      lowercase(opcode);

      // Macro definitions are skipped in second pass
      //
      if (inmacro) {
        if (!strcmp(opcode, "endm")) { voc=1; inmacro = 0; }
        continue;
      }

      if ((isalnum(label[0])) || (label[0] == '_') || (label[0] == '@')) {

        // This should never be true unless there is a bug in the assember
        //
        if (strcmp(opcode, "equ")) {
          if (symLookup(label) != lc) {
            message(MSG_CRITICAL, "Second pass value mismatch on label '%s'. Assembler error?",
                label);
          }
          if (lstp) fprintf(lstp, "%04x\t\t%s:\n", lc, label);
          bsr = -1;  // Invalidate BSR tracking on label
        }
      }


      if (opcode[0] != '\0') {
      
        // fprintf(stderr, "Processing opcode: '%s'\n", opcode);

        if (try_opcode(opcode, operands, comment) < 0) {
          voc = 0;
          if (!strcmp(opcode, "end")) {
            message(MSG_INFO, "%04x END statement reached.", lc);
            // Close all input files
            while ((sfp = pop_source(sfp)));
            // Break out of the loops
            done = 1;
            break;
          }
          if (!strcmp(opcode, "equ")) voc = 1;
          if (!strcmp(opcode, "include")) {
            voc = 1;
            sfp = do_include(sfp, operands);
          }
          if (!strcmp(opcode, "banksel")) { voc=1; do_banksel(operands); }
          if (!strcmp(opcode, "macro")) { voc=1; inmacro = 1; }
          if (!strcmp(opcode, "endm")) { voc=1; inmacro = 0; }
          if (!strcmp(opcode, "org")) { voc=1; do_org(operands); }
          if (!strcmp(opcode, "__config1")) { voc=1; do_config(1, operands); }
          if (!strcmp(opcode, "__config2")) { voc=1; do_config(2, operands); }
          if (macInvocations(opcode, 0) >= 0) {
            voc=1;
            do_macro(opcode, operands);
          }
          if (!voc) message(MSG_WARNING, "%04x: Invalid opcode or directive '%s'", lc, opcode);
        } else {
          // Opcode found. Normally do nothing, unless BSR changed.
          // Right now we're just invalidating BSR tracking.  We could
          //   read the operand and update 'bsr' accordingly.
          // if (!strcmp(opcode, "movlb")) bsr = -1;
        }

      }

      // printf("%s", ibuf);
    }
  } while (!done && (sfp = pop_source(sfp)));

  // Last sfp will have been closed by pop_source.
  // fclose(sfp);

  fclose(lstp);

  fprintf(stderr, "\n");
  sprintf(ibuf, "%s.hex", fnroot);
  mmWriteHexfile(chipmem, ibuf);

  sprintf(ibuf, "%s.sym", fnroot);
  if ((symp = fopen(ibuf, "w")) != NULL) {
    symDump(symp);
    fclose(symp);
  }


  symDestroy();
  macDestroy();
  ocitDestroy();

  mmFree(chipmem);

  exit(0);
}


